///////////////////////////////////////////////////////////////////////////////
///
/// Copyright (c) 2007 - Oliver Schneider (assarbad.net)
///
/// Original filename: cmdassys.c
/// Project          : cmdassys
/// Date of creation : 2007-04-08
/// Author(s)        : 
///
/// Purpose          : Starts a command prompt as SYSTEM
///
/// Revisions:
///  0000 [2007-04-08] Initial revision.
///
///////////////////////////////////////////////////////////////////////////////

// $Id$

///////////////////////////////////////////////////////////////////////////////
// #define UNICODE
// #define _UNICODE
// These two defines are given implicitly through the settings of C_DEFINES in
// the SOURCES file of the project. Hence change them there and there only.
///////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <tchar.h>
#include "execommon.h"
#include "exeversion.h"
#include <Windows.h>

namespace
{
	SERVICE_STATUS_HANDLE s_hss = NULL;
	SERVICE_STATUS s_ss = { SERVICE_WIN32_OWN_PROCESS, SERVICE_RUNNING, 0, 0, 0, 0, 0 };
	const LPTSTR lpcsServiceName = TEXT("CMD_AS_SYSTEM");
	const LPTSTR lpcsDefaultCommandLine = TEXT("cmd.exe /X /K mode con cols=80 lines=50 & cls");
}

#if DBG
#include <stdarg.h>
void DbgPrint(PCHAR FmtString, ...)
{
   va_list argptr;
   va_start(argptr, FmtString);
   vprintf(FmtString, argptr);
   printf("\n");
   if(FILE* file = fopen("C:\\logfile.txt", "a+"))
   {
	   vfprintf(file, FmtString, argptr);
	   fprintf(file, "\n");
	   fclose(file);
   }
}
#else
#define DbgPrint /##/
#endif

bool IsAdmin()
{
	HANDLE hToken = NULL;
	if(OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken))
	__try
	{
		DWORD cBytes = 0;
		PTOKEN_GROUPS pTokenGroups = NULL;
		// Retrieve the number of bytes required and allocate some memory
		if((!GetTokenInformation(hToken, TokenGroups, NULL, 0, &cBytes)) &&
			(pTokenGroups = PTOKEN_GROUPS(GlobalAlloc(GPTR, cBytes))) &&
			(GetTokenInformation(hToken, TokenGroups, pTokenGroups, cBytes, &cBytes))
			)
		__try
		{
			PSID pSid = NULL;
			SID_IDENTIFIER_AUTHORITY auth = SECURITY_NT_AUTHORITY;
			// Allocate the SID for the administrator group
			if(
				AllocateAndInitializeSid(&auth, 2, SECURITY_BUILTIN_DOMAIN_RID, DOMAIN_ALIAS_RID_ADMINS, 0, 0, 0, 0, 0, 0, &pSid) &&
				pSid
				)
			__try
			{
				// Loop through all groups
				for(DWORD i = 0; i < pTokenGroups->GroupCount; i++)
				{
					// If group membership is found, bail out
					if(FALSE != EqualSid(pTokenGroups->Groups[i].Sid, pSid))
					{
						return true;
					}
				}
			}
			__finally
			{
				FreeSid(pSid);
			}
		}
		__finally
		{
			GlobalFree(HGLOBAL(pTokenGroups));
		}
	}
	__finally
	{
		CloseHandle(hToken);
	}
	return false;
}

bool IsLocalSystem()
{
	HANDLE hToken = NULL;
	if(OpenProcessToken(GetCurrentProcess(), TOKEN_QUERY, &hToken))
	__try
	{
		DWORD cBytes = 0;
		PTOKEN_USER pTokenUser = NULL;
		// Retrieve the number of bytes required and allocate some memory
		if((!GetTokenInformation(hToken, TokenUser, NULL, 0, &cBytes)) &&
			(pTokenUser = PTOKEN_USER(GlobalAlloc(GPTR, cBytes))) &&
			(GetTokenInformation(hToken, TokenUser, pTokenUser, cBytes, &cBytes))
			)
		__try
		{
			PSID pSid = NULL;
			SID_IDENTIFIER_AUTHORITY auth = SECURITY_NT_AUTHORITY;
			// Allocate the SID for the SYSTEM 'user'
			if(
				(AllocateAndInitializeSid(&auth, 1, SECURITY_LOCAL_SYSTEM_RID, 0, 0, 0, 0, 0, 0, 0, &pSid)) &&
				pSid
				)
			__try
			{
				return (FALSE != EqualSid(pTokenUser->User.Sid, pSid));
			}
			__finally
			{
				FreeSid(pSid);
			}
		}
		__finally
		{
			GlobalFree(HGLOBAL(pTokenUser));
		}
	}
	__finally
	{
		CloseHandle(hToken);
	}
	return false;
}

LPTSTR __forceinline MyAllocString(SIZE_T Bytes)
{
	return LPTSTR(GlobalAlloc(GPTR, Bytes));
}

bool __forceinline MyFreeString(LPCTSTR String)
{
	return (String) ? true : (FALSE != GlobalFree(HGLOBAL(String)));
}

LPTSTR MyGetModuleName()
{
	DWORD cchLen = (MAX_PATH + 1);
	if(LPTSTR lpcsResult = MyAllocString(sizeof(TCHAR) * cchLen))
	{
		if(GetModuleFileName(NULL, lpcsResult, cchLen))
		{
			DbgPrint("MyGetModuleName() -> %ws", lpcsResult);
			return lpcsResult;
		}
		else
		{
			MyFreeString(lpcsResult);
		}
	}
	return NULL;
}

LPTSTR MyGetCurrentDirectory()
{
	DWORD cchLen = (MAX_PATH + 1);
	if(LPTSTR lpcsResult = MyAllocString(sizeof(TCHAR) * cchLen))
	{
		if(GetCurrentDirectory(cchLen, lpcsResult))
		{
			DbgPrint("MyGetCurrentDirectory() -> %ws", lpcsResult);
			return lpcsResult;
		}
		else
		{
			MyFreeString(lpcsResult);
		}
	}
	return NULL;
}

LPTSTR MyGetUserName()
{
	DWORD cchLen = 0;
	if(!GetUserName(NULL, &cchLen))
	{
		++cchLen;
		if(LPTSTR lpcsResult = MyAllocString(sizeof(TCHAR) * cchLen))
		{
			if(GetUserName(lpcsResult, &cchLen))
			{
				DbgPrint("MyGetUserName() -> %ws", lpcsResult);
				return lpcsResult;
			}
			else
			{
				MyFreeString(lpcsResult);
			}
		}
	}
	return NULL;
}

LPTSTR MyGetComputerName()
{
	DWORD cchLen = 0;
	if(!GetComputerName(NULL, &cchLen))
	{
		++cchLen;
		if(LPTSTR lpcsResult = MyAllocString(sizeof(TCHAR) * cchLen))
		{
			memset(lpcsResult, 0, sizeof(TCHAR) * cchLen);
			if(GetComputerName(lpcsResult, &cchLen))
			{
				DbgPrint("MyGetComputerName() -> %ws", lpcsResult);
				return lpcsResult;
			}
			else
			{
				MyFreeString(lpcsResult);
			}
		}
	}
	return NULL;
}

LPTSTR MyDesktopName()
{
	// Connect to the desktop (needed on Windows NT 3.51)
	GetDesktopWindow();
	HANDLE hDesktop = GetThreadDesktop(GetCurrentThreadId());
	HANDLE hWinSta = GetProcessWindowStation();
	if((hDesktop) && (hWinSta))
	{
		DWORD cchNeeded = (MAX_PATH + 1);
		if(LPTSTR lpcsDesktopName = MyAllocString(sizeof(TCHAR) * cchNeeded))
		__try
		{
			DWORD temp = 0;
			if(GetUserObjectInformation(hDesktop, UOI_NAME, PVOID(lpcsDesktopName), sizeof(TCHAR) * cchNeeded, &temp))
			{
				if(LPTSTR lpcsWinStaName = MyAllocString(sizeof(TCHAR) * cchNeeded))
				__try
				{
					temp = 0;
					if(GetUserObjectInformation(hWinSta, UOI_NAME, PVOID(lpcsWinStaName), sizeof(TCHAR) * cchNeeded, &temp))
					{
						cchNeeded = 2 + _tcslen(lpcsDesktopName) + _tcslen(lpcsWinStaName);
						if(LPTSTR lpcsResult = MyAllocString(sizeof(TCHAR) * cchNeeded))
						{
							// Concatenate the names
							_tcscpy(lpcsResult, lpcsWinStaName);
							_tcscat(lpcsResult, TEXT("\\"));
							_tcscat(lpcsResult, lpcsDesktopName);
							return lpcsResult;
						}
					}
				}
				__finally
				{
					MyFreeString(lpcsWinStaName);
				}
			}
		}
		__finally
		{
			MyFreeString(lpcsDesktopName);
		}
	}
	return NULL;
}

BOOL MyCreateProcess(LPTSTR CommandLine, LPTSTR DesktopName, LPTSTR CurrentDirectory)
{
	const LPTSTR csConsoleTitle = TEXT("CMD.EXE running as %s@%s\\%s");
	LPTSTR lpcsUserName = MyGetUserName();
	LPTSTR lpcsComputerName = MyGetComputerName();
	DWORD dwLen = sizeof(TCHAR) * (
								_tcslen(csConsoleTitle) +
								((DesktopName) ? _tcslen(DesktopName) : 10) +
								((lpcsUserName) ? _tcslen(lpcsUserName) : 10) +
								((lpcsComputerName) ? _tcslen(lpcsComputerName) : 10)
								);
	LPTSTR lpcsConTitle = NULL;
	__try
	{
		if(lpcsConTitle = MyAllocString(dwLen))
		{
			_stprintf(lpcsConTitle, csConsoleTitle, lpcsUserName, lpcsComputerName, DesktopName);
		}
		STARTUPINFO si = {
			sizeof(STARTUPINFO),
			NULL,
			DesktopName,
			lpcsConTitle, // Console title
			CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT, CW_USEDEFAULT,
			80, 50,
			FOREGROUND_GREEN | FOREGROUND_INTENSITY,
			STARTF_USECOUNTCHARS | STARTF_USEFILLATTRIBUTE | STARTF_USEPOSITION,
			0, 0, 0, 0, 0, 0
		};

		PROCESS_INFORMATION pi  = {0, 0, 0, 0};
		if(CreateProcess(NULL, CommandLine, NULL, NULL, FALSE, CREATE_NEW_CONSOLE |
						CREATE_NEW_PROCESS_GROUP, NULL, NULL, &si, &pi))
		{
			DbgPrint("CreateProcess succeeded!");
			CloseHandle(pi.hThread);
			CloseHandle(pi.hProcess);
		}
		DbgPrint("CreateProcess %d == %8.8X", GetLastError(), GetLastError());
	}
	__finally
	{
		if(lpcsUserName) { MyFreeString(lpcsUserName); }
		if(lpcsComputerName) { MyFreeString(lpcsComputerName); }
		if(lpcsConTitle) { MyFreeString(lpcsConTitle); }
	}
	return TRUE;
}

DWORD StartAsService(LPTSTR lpcsCommandLine)
{
	if(SC_HANDLE hSCM = OpenSCManager(NULL, SERVICES_ACTIVE_DATABASE, SC_MANAGER_CREATE_SERVICE))
	__try
	{
		if(LPTSTR lpcsSvcPath = MyGetModuleName())
		__try
		{
			const DWORD cMaxRetries = 3;
			for(DWORD i = 0; i < cMaxRetries; i++)
			{
				if(SC_HANDLE hSvc = CreateService(hSCM, lpcsServiceName, lpcsServiceName,
					SERVICE_START | SERVICE_QUERY_STATUS | DELETE,
					SERVICE_WIN32_OWN_PROCESS, SERVICE_DEMAND_START,
					SERVICE_ERROR_NORMAL, lpcsSvcPath, NULL, NULL,
					NULL, NULL, NULL))
				__try
				{
					if(LPTSTR lpcsDesktop = MyDesktopName())
					__try
					{
						if(LPTSTR lpcsCurrDir = MyGetCurrentDirectory())
						__try
						{
							const DWORD cArguments = 3;
							// Find the last backslash and replace it by a terminating zero!
							LPTSTR SvcArguments[cArguments] = {lpcsDesktop, lpcsCurrDir, (lpcsCommandLine) ? lpcsCommandLine : lpcsDefaultCommandLine };
							if(StartService(hSvc, cArguments, (LPCTSTR*)(SvcArguments)))
							{
								DeleteService(hSvc);
							}
							return ERROR_SUCCESS;
						}
						__finally
						{
							MyFreeString(lpcsCurrDir);
						}
					}
					__finally
					{
						MyFreeString(lpcsDesktop);
					}
				}
				__finally
				{
					CloseServiceHandle(hSvc);
				}
				else
				{
					if(ERROR_SERVICE_EXISTS == GetLastError())
					{
						if(SC_HANDLE hSvc = OpenService(hSCM, lpcsServiceName, DELETE))
						__try
						{
							DeleteService(hSvc);
						}
						__finally
						{
							CloseServiceHandle(hSvc);
						}
					}
				}
			}
		}
		__finally
		{
			MyFreeString(lpcsSvcPath);
		}
	}
	__finally
	{
		CloseServiceHandle(hSCM);
	}
	return ERROR_FUNCTION_FAILED;
}

VOID WINAPI ServiceHandlerFunc(DWORD dwControl)
{
	DbgPrint("ServiceHandlerFunc");
	SetServiceStatus(s_hss, &s_ss);
}

VOID WINAPI ServiceMainFunc(DWORD dwNumServicesArgs, LPTSTR *lpServiceArgVectors)
{
	s_hss = RegisterServiceCtrlHandler(lpcsServiceName, ServiceHandlerFunc);
	SetServiceStatus(s_hss, &s_ss);

	DbgPrint("Now creating process #0=%ws, #1=%ws, #2=%ws, #3=%ws", lpServiceArgVectors[0], lpServiceArgVectors[1], lpServiceArgVectors[2], lpServiceArgVectors[3]);
	MyCreateProcess(lpServiceArgVectors[3], lpServiceArgVectors[1], lpServiceArgVectors[2]);

	s_ss.dwCurrentState = SERVICE_STOPPED;
	SetServiceStatus(s_hss, &s_ss);
}

int __cdecl main(int argc, _TCHAR *argv[])
{
	if(GetVersion() & 0x80000000)
	{
// 		_tprintf(TEXT("This program can only be run on the NT platform!\n"));
		return -1;
	}
	DbgPrint("Started");
	// Connect to the desktop (needed on Windows NT 3.51)
	GetDesktopWindow();
	// We restrict the program to admin, only admins have the required privileges
	// by default anyway.
	if(IsAdmin())
	{
		DbgPrint("Started as admin");
		// If we run under the SYSTEM account assume we run as a service
		if(IsLocalSystem())
		{
			DbgPrint("Passing control to SC dispatcher");
			SERVICE_TABLE_ENTRY ste[2] = {{lpcsServiceName, ServiceMainFunc}, {NULL, NULL}};
			// Don't care about success/failure
			StartServiceCtrlDispatcher(&ste[0]);
		}
		else
		{
			DbgPrint("Starting service");
			WNDCLASS wc = { 0, 0, 0, 0, GetModuleHandle(NULL), 0, 0, 0, 0, TEXT("CMDasSYS_NonsenseWndClass") };
			RegisterClass(&wc);
			// Create and start the service, then quit
			StartAsService(NULL);
		}
	}
}
